import os
import unittest
import pandas as pd
from propertime.utils import dt, s_from_dt

from ..utils import detect_encoding, detect_periodicity, detect_sampling_interval
from ..utils import _compute_coverage, _compute_data_loss, _compute_validity_regions
from ..utils import _Gaussian
from ..utils import rescale
from ..utils import _is_index_based, _is_key_based, _has_numerical_values
from ..datastructures import DataTimePoint, TimeSeries
from ..storages import CSVFileStorage
from ..units import TimeUnit

# Setup logging
from .. import logger
logger.setup()

# Support function for attaching validity regions to a series
def attach_validity_regions(series, sampling_interval=None):
    validity_regions = _compute_validity_regions(series, sampling_interval=sampling_interval)
    for point in series:
        point.valid_from = validity_regions[point.t][0]
        point.valid_to = validity_regions[point.t][1]

# Set test data path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'


class TestComputeValidityRegions(unittest.TestCase):

    def test_standard(self):

        series = TimeSeries()
        series.append(DataTimePoint(t = 7, data = {'value': 7}))
        series.append(DataTimePoint(t = 12, data = {'value': 12}))
        series.append(DataTimePoint(t = 17, data = {'value': 17}))
        series.append(DataTimePoint(t = 22, data = {'value': 22}))
        series.append(DataTimePoint(t = 27, data = {'value': 27}))

        # All of them
        expected_results = {7: [4.5, 9.5], 12: [9.5, 14.5], 17: [14.5, 19.5], 22: [19.5, 24.5], 27: [24.5, 29.5]}
        results = _compute_validity_regions(series)
        self.assertEqual(results, expected_results)

        # Only from-to
        expected_results = {12: [9.5, 14.5], 17: [14.5, 19.5], 22: [19.5, 24.5]}
        results = _compute_validity_regions(series, from_t=10, to_t=20)
        self.assertEqual(results, expected_results)

        # Shrink them according to the from-to
        expected_results = {12: [10, 14.5], 17: [14.5, 19.5], 22: [19.5, 20]}
        results = _compute_validity_regions(series, from_t=10, to_t=20, cut=True)
        self.assertEqual(results, expected_results)

        # Force a specific sampling intervals
        expected_results = {12: [11.5, 12.5], 17: [16.5, 17.5]}
        results = _compute_validity_regions(series, from_t=10, to_t=20, sampling_interval=1)
        self.assertEqual(results, expected_results)

        # Single-element series
        single_element_series = TimeSeries()
        single_element_series.append(DataTimePoint(t = 40, data = {'value': 4571.55}))
        expected_results = {40: [35.0, 45.0]}

        # The series has only one element and no sampling_interval is provided, no idea how to compute validity
        with self.assertRaises(ValueError):
            _compute_validity_regions(series = single_element_series, from_t = 30, to_t = 60)

        # Call by providing the sampling_interval
        results = _compute_validity_regions(series = single_element_series, from_t = 30, to_t = 60, sampling_interval=10)

        self.assertEqual(results, expected_results)


    def test_prev_next_points(self):
        series = TimeSeries()
        series.append(DataTimePoint(t = 7, data = {'value': 7}))
        series.append(DataTimePoint(t = 12, data = {'value': 12}))
        series.append(DataTimePoint(t = 17, data = {'value': 17}))
        series.append(DataTimePoint(t = 22, data = {'value': 22}))
        series.append(DataTimePoint(t = 27, data = {'value': 27}))

        # Test prev-next
        expected_results = {7: [4.5, 9.5], 12: [9.5, 14.5], 17: [14.5, 19.5], 22: [19.5, 24.5], 27: [24.5, 29.5]}
        results = _compute_validity_regions(series, from_t=9, to_t=25)
        self.assertEqual(results, expected_results)

        # Test cutted prev-next
        expected_results = {7: [9, 9.5], 12: [9.5, 14.5], 17: [14.5, 19.5], 22: [19.5, 24.5], 27: [24.5, 25]}
        results = _compute_validity_regions(series, from_t=9, to_t=25, cut=True)
        self.assertEqual(results, expected_results)


    def test_overlaps(self):

        series = TimeSeries()
        series.append(DataTimePoint(t = 7, data = {'value': 7}))
        series.append(DataTimePoint(t = 12, data = {'value': 12}))
        series.append(DataTimePoint(t = 13, data = {'value': 13}))
        series.append(DataTimePoint(t = 17, data = {'value': 17}))
        series.append(DataTimePoint(t = 22, data = {'value': 22}))
        series.append(DataTimePoint(t = 27, data = {'value': 27}))

        expected_results = {7: [4.5, 9.5], 12: [9.5, 12.5], 13: [12.5, 15], 17: [15, 19.5], 22: [19.5, 24.5], 27: [24.5, 29.5]}
        results = _compute_validity_regions(series)
        self.assertEqual(results, expected_results)


    def test_major_overlaps(self):

        series = TimeSeries()
        series.append(DataTimePoint(t = 7, data = {'value': 7}))
        series.append(DataTimePoint(t = 12, data = {'value': 12}))
        series.append(DataTimePoint(t = 12.1, data = {'value': 12.1}))
        series.append(DataTimePoint(t = 12.3, data = {'value': 12.3}))
        series.append(DataTimePoint(t = 13, data = {'value': 13}))
        series.append(DataTimePoint(t = 17, data = {'value': 17}))
        series.append(DataTimePoint(t = 22, data = {'value': 22}))
        series.append(DataTimePoint(t = 27, data = {'value': 27}))

        expected_results = {7: [4.5, 9.5], 12: [9.5, 12.05], 12.1: [12.05, 12.2], 12.3: [12.2, 12.65], 13: [12.65, 15.0], 17: [15.0, 19.5], 22: [19.5, 24.5], 27: [24.5, 29.5]}
        results = _compute_validity_regions(series)
        self.assertEqual(results, expected_results)



class TestComputeCoverageAndDataLoss(unittest.TestCase):

    def setUp(self):

        # All the following time series have point with validity=1m

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        self.timeseries_1 = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            datatimepoint = DataTimePoint(t = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.timeseries_1.append(datatimepoint)
        attach_validity_regions(self.timeseries_1)

        # Time series from 17:00:00 to 17:30:00 (Europe/Rome)
        self.timeseries_2 = TimeSeries()
        start_t = 1436022000
        for i in range(34):
            datatimepoint = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.timeseries_2.append(datatimepoint)
        attach_validity_regions(self.timeseries_2)

        # Time series from 17:00:00 to 17:20:00 (Europe/Rome)
        self.timeseries_3 = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(23):
            datatimepoint = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.timeseries_3.append(datatimepoint)
        attach_validity_regions(self.timeseries_3)

        # Time series from 17:10:00 to 17:30:00 (Europe/Rome)
        self.timeseries_4 = TimeSeries()
        start_t = 1436022000 + 600
        for i in range(21):
            datatimepoint = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.timeseries_4.append(datatimepoint)
        attach_validity_regions(self.timeseries_4)

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        self.timeseries_5 = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i <21:
                continue
            datatimepoint = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.timeseries_5.append(datatimepoint)
        attach_validity_regions(self.timeseries_5)


        # The following time series has point with validity=15m

        # Time series from 2019,10,1,1,0,0 to 2019,10,1,6,0,0 (Europe/Rome)
        from_dt  = dt(2019,10,1,1,0,0, tz='Europe/Rome')
        to_dt    = dt(2019,10,1,6,0,0, tz='Europe/Rome')
        time_unit = TimeUnit('15m')
        self.timeseries_6 = TimeSeries()
        slider_dt = from_dt
        count = 0
        while slider_dt < to_dt:
            if count not in [1, 6, 7, 8, 9, 10]:
                datatimepoint = DataTimePoint(t    = s_from_dt(slider_dt),
                                                data = {'temperature': 154+count})
                self.timeseries_6.append(datatimepoint)
            slider_dt = slider_dt + time_unit
            count += 1
        attach_validity_regions(self.timeseries_6)


    def test_compute_coverage(self):

        from_t = 1436022000       # 2015-07-04 17:00:00+02:00
        to_t   = 1436022000+1800  # 2015-07-04 17:30:00+02:00

        # A) Full coverage (coverage=1.0) and again, to test "reproducibility"
        coverage = _compute_coverage(series = self.timeseries_1, from_t = from_t, to_t = to_t)
        self.assertEqual(coverage, 1.0)
        coverage = _compute_coverage(series = self.timeseries_1, from_t = from_t, to_t = to_t)
        self.assertEqual(coverage, 1.0)


        # B) Full coverage (coverage=1.0) without prev/next in the time series
        coverage = _compute_coverage(series = self.timeseries_1, from_t = from_t, to_t = to_t)
        self.assertEqual(coverage, 1.0)
        self.assertEqual(coverage, 1.0)


        # C) Missing ten minutes over 30 at the end (coverage=0.683))
        coverage = _compute_coverage(series = self.timeseries_3, from_t = from_t, to_t = to_t)

        # 20 minutes plus other 30 secs validity for the 20th point over 30 minutes
        self.assertEqual(coverage, ( ((20*60.0)+30.0) / (30*60.0)) )


        # D) Missing ten minutes over 30 at the beginning (coverage=0.683)
        coverage = _compute_coverage(series = self.timeseries_4, from_t = from_t, to_t = to_t)

        # 20 minutes plus other 30 secs (previous) validity for the 10th point over 30 minutes
        self.assertEqual(coverage, ( ((20*60.0)+30.0) / (30*60.0)) )


        # E) Missing eleven minutes over 30 in the middle (coverage=0.66)
        coverage = _compute_coverage(series = self.timeseries_5, from_t = from_t, to_t = to_t)

        # 20 minutes plus other 30 secs (previous) validity for the 10th point over 30 minutes
        self.assertAlmostEqual(coverage, (2.0/3.0))


        # F) Missing half slot before slot re-start
        from_t = s_from_dt(dt=dt(2019,10,1,3,30,0, tz='Europe/Rome'))
        to_t   = s_from_dt(dt=dt(2019,10,1,3,45,0, tz='Europe/Rome'))
        coverage = _compute_coverage(series = self.timeseries_6, from_t = from_t, to_t = to_t)

        self.assertAlmostEqual(coverage, (0.5))

        # G) Border conditions for from_t and to_t:
        timeseries = TimeSeries()
        timeseries.append(DataTimePoint(t = 20, data = {'temperature': 23}))
        timeseries.append(DataTimePoint(t = 30, data = {'temperature': 23}))
        timeseries.append(DataTimePoint(t = 40, data = {'temperature': 23}))
        timeseries.append(DataTimePoint(t = 60, data = {'temperature': 23}))
        timeseries.append(DataTimePoint(t = 70, data = {'temperature': 23}))
        attach_validity_regions(timeseries)

        coverage = _compute_coverage(series = timeseries, from_t = 30, to_t = 60)
        self.assertAlmostEqual(coverage, (2.0/3.0))

        # H) single-element series (requires providing the sampling_interval when computing the validity regions)
        single_element_series = TimeSeries()
        single_element_series.append(DataTimePoint(t = 40, data = {'temperature': 23}))

        attach_validity_regions(single_element_series, sampling_interval=10)
        coverage = _compute_coverage(series = single_element_series, from_t = 30, to_t = 60)

        # 30 to 60, point is from 35 to 45 -> 10 seconds out of 30
        self.assertAlmostEqual(coverage, (1.0/3.0))


    def test_compute_data_loss(self):

        # Basic series
        series = TimeSeries()
        series.append(DataTimePoint(t = 20, data = {'temperature': 23}))
        series.append(DataTimePoint(t = 30, data = {'temperature': 23}))
        series.append(DataTimePoint(t = 40, data = {'temperature': 23}))
        series.append(DataTimePoint(t = 60, data = {'temperature': 23}))
        series.append(DataTimePoint(t = 70, data = {'temperature': 23}))
        attach_validity_regions(series)

        # Test basic
        self.assertAlmostEqual(_compute_data_loss(series, from_t = 45, to_t = 55, sampling_interval=10), 1.0)
        self.assertAlmostEqual(_compute_data_loss(series, from_t = 50, to_t = 60, sampling_interval=10), 0.5)
        self.assertAlmostEqual(_compute_data_loss(series, from_t = 65, to_t = 75, sampling_interval=10), 0.0)

        # Test out of boundaries
        data_loss = _compute_data_loss(series, from_t = 100, to_t = 120, sampling_interval=10)
        self.assertEqual(data_loss, 1)

        # Test for 1-element series (requires providing the sampling_interval when computing the validity regions)
        single_element_series = TimeSeries()
        single_element_series.append(DataTimePoint(t = 40, data = {'temperature': 23}))
        attach_validity_regions(single_element_series, sampling_interval=10)

        # Only 10 seconds out of 30 (point has validity regions 35-45, interval is from 30 to 60)
        data_loss = _compute_data_loss(single_element_series, from_t = 30, to_t = 60, sampling_interval=10)
        self.assertAlmostEqual(data_loss, (2.0/3.0))

        # Test two elements, some border condition for sull fata losses may arise
        series_two_elements = TimeSeries()
        series_two_elements.append(DataTimePoint(t = 40, data = {'temperature': 23}))
        series_two_elements.append(DataTimePoint(t = 60, data = {'temperature': 23}))
        attach_validity_regions(series_two_elements, sampling_interval=10)

        self.assertEqual(_compute_data_loss(series_two_elements, from_t = 45, to_t = 55, sampling_interval=10), 1)

        # Series with pre-existent data losses, which have to be taken into account as well
        series_with_data_losses = TimeSeries()
        series_with_data_losses.append(DataTimePoint(t = 20, data = {'temperature': 23}, data_loss=0))
        series_with_data_losses.append(DataTimePoint(t = 30, data = {'temperature': 23}, data_loss=0))
        series_with_data_losses.append(DataTimePoint(t = 40, data = {'temperature': 23}, data_loss=0.5))
        series_with_data_losses.append(DataTimePoint(t = 60, data = {'temperature': 23}, data_loss=0))
        series_with_data_losses.append(DataTimePoint(t = 70, data = {'temperature': 23}, data_loss=0))
        attach_validity_regions(series_with_data_losses)

        # Test missing + already ( 0 + 0.5 + 1 + 0 = 1.5 / 3 = 0.5)
        data_loss = _compute_data_loss(series_with_data_losses, from_t = 30, to_t = 60, sampling_interval=10, force=True)
        self.assertAlmostEqual(data_loss, 0.5)

        # Test also on actually resampled series
        resampled_series = series.resample(10)
        attach_validity_regions(resampled_series)

        data_loss = _compute_data_loss(resampled_series, from_t = 30, to_t = 60)
        self.assertAlmostEqual(data_loss, 1.0/3.0)

        resampled_series_5 = self.timeseries_5.resample(60)
        attach_validity_regions(resampled_series_5)

        data_loss = _compute_data_loss(resampled_series_5, from_t = 1436022300, to_t = 1436022600)
        self.assertAlmostEqual(data_loss, 0.3)

        # TODO: This test was introduced after finding a bug, and needs to be better unrolled.
        # the bug was raising when there were overlapping points at the start or end of the slot.
        from timeseria import  storages
        csv_storage = storages.CSVFileStorage(TEST_DATA_PATH + 'csv/temperature.csv')
        series = csv_storage.get(limit=400)
        series = series[200:300]
        resampled_series = series.resample(600)
        attach_validity_regions(resampled_series)
        data_loss = _compute_data_loss(resampled_series, from_t = 1546700400, to_t = 1546711200)
        self.assertAlmostEqual(data_loss, 1.0)


class TestGetPeriodicity(unittest.TestCase):

    def test_get_periodicity(self):

        univariate_timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get()

        univariate_1h_timeseries = univariate_timeseries.aggregate('1h')

        perdiodicity = detect_periodicity(univariate_1h_timeseries)

        self.assertEqual(perdiodicity, 24)


class TestDetectSamplingInterval(unittest.TestCase):

    def test_detect_sampling_interval(self):

        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/humitemp_short.csv').get()
        self.assertEqual(detect_sampling_interval(timeseries), 61)
        self.assertAlmostEqual(detect_sampling_interval(timeseries, confidence=True)[1], 0.7368421)

        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/shampoo_sales.csv', date_column = 'Month', date_format = '%y-%m').get()
        self.assertEqual(detect_sampling_interval(timeseries), 2678400)  # 2678400/60/60/24 = 31 Days (the most frequent)

        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temperature.csv').get()
        self.assertEqual(detect_sampling_interval(timeseries), 600)

        timeseries = CSVFileStorage(TEST_DATA_PATH + '/csv/temp_short_1h.csv').get()
        self.assertEqual(detect_sampling_interval(timeseries), 3600)


class TestDetectEncoding(unittest.TestCase):

    def test_detect_encoding(self):

        encoding = detect_encoding('{}/csv/shampoo_sales.csv'.format(TEST_DATA_PATH), streaming=False)
        self.assertEqual(encoding, 'ascii')


class TestGaussian(unittest.TestCase):

    def test_gaussian(self):

        test_data = [1,3,7,8,8,8,9,9,9,9,9,10,10,10,10,10,10,11,11,11,11,12,12,12,13,13,15,15,18,20]

        gaussian = _Gaussian.from_data(test_data)

        # Check parameters
        self.assertAlmostEqual(gaussian.mu, 10.466, places=2)
        self.assertAlmostEqual(gaussian.sigma, 3.639, places=2)

        # Check callable function behavior
        self.assertAlmostEqual(gaussian(gaussian.mu), 0.1096, places=3)

        self.assertAlmostEqual(gaussian(10), 0.1087, places=4)

        # Check solver (find xes):
        self.assertAlmostEqual(gaussian.find_xes(y=0.1)[0], 8.9079, places=3)
        self.assertAlmostEqual(gaussian.find_xes(y=0.1)[1], 12.0253, places=3)
        self.assertAlmostEqual(gaussian(gaussian.find_xes(y=0.05)[0]), 0.05, places=3)
        self.assertAlmostEqual(gaussian(gaussian.find_xes(y=0.05)[1]), 0.05, places=3)

        # Edge case to find back the peak
        self.assertAlmostEqual(gaussian.find_xes(y=0.1096)[0], 10.4422, places=2)
        self.assertAlmostEqual(gaussian.find_xes(y=0.1096)[1], 10.4910, places=2)

        # Test the CDF now
        self.assertEqual(gaussian.cumulative(gaussian.mu), 0.5)


class TestRescale(unittest.TestCase):

    def test_rescale(self):

        with self.assertRaises(ValueError):
            rescale(value=0.5, source_from=1, source_to=2, target_from=1, target_to=0)
        with self.assertRaises(ValueError):
            rescale(value=2.5, source_from=1, source_to=2, target_from=1, target_to=0)

        self.assertEqual(rescale(value=7.5, source_from=5, source_to=10, target_from=0, target_to=1),0.5)
        self.assertEqual(rescale(value=7.5, source_from=5, source_to=10, target_from=0, target_to=2),1)
        self.assertEqual(rescale(value=7.5, source_from=5, source_to=10, target_from=1, target_to=2),1.5)
        self.assertEqual(rescale(value=7.5, source_from=5, source_to=10, target_from=10, target_to=20),15)
        self.assertEqual(rescale(value=0.5, source_from=0, source_to=1, target_from=7, target_to=8),7.5)

        self.assertAlmostEqual(rescale(value=0.98, source_from=0.000016, source_to=1.0, target_from=0, target_to=1), 0.97999967)


class TestDetectDataTypes(unittest.TestCase):

    def test_detect_data_types(self):

        data = [1,2,3]
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), True)

        data = [1,'z',3]
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), False)

        data = [1,2,None]
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), False)


        data = (1,2,3)
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), True)

        data = 'hello'
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), False)

        data = {'a':1, 'b':2}
        self.assertEqual(_is_index_based(data), False)
        self.assertEqual(_is_key_based(data), True)
        self.assertEqual(_has_numerical_values(data), True)

        data = {'a':1, 'b':'z'}
        self.assertEqual(_is_index_based(data), False)
        self.assertEqual(_is_key_based(data), True)
        self.assertEqual(_has_numerical_values(data), False)

        data = {'a':1, 'b':None}
        self.assertEqual(_is_index_based(data), False)
        self.assertEqual(_is_key_based(data), True)
        self.assertEqual(_has_numerical_values(data), False)

        data = pd.Series([4,5,6])
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), True)

        data = pd.Series([4,5,'z'])
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), False)

        data = pd.Series([4,5,6])
        data.index = ['a','b','c']
        self.assertEqual(_is_index_based(data), True)
        self.assertEqual(_is_key_based(data), False)
        self.assertEqual(_has_numerical_values(data), True)




import unittest
import os
from propertime.utils import dt, s_from_dt, dt_from_str

from ..datastructures import DataTimePoint, TimeSeries
from ..transformations import Resampler, Aggregator, _TimeSeriesDenseView
from ..units import TimeUnit

# Setup logging
from .. import logger
logger.setup()

# Set test data path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'


class TestTimeSeriesDenseView(unittest.TestCase):

    def test_dense_TimeSeriesDenseView(self):
        series = TimeSeries()
        series.append(DataTimePoint(t = 0, data = {'value': 0}))
        series.append(DataTimePoint(t = 1, data = {'value': 1}))
        series.append(DataTimePoint(t = 2, data = {'value': 2}))
        series.append(DataTimePoint(t = 3, data = {'value': 3}))
        series.append(DataTimePoint(t = 4, data = {'value': 4}))
        series.append(DataTimePoint(t = 7, data = {'value': 7}))
        series.append(DataTimePoint(t = 8, data = {'value': 8}))
        series.append(DataTimePoint(t = 9, data = {'value': 9}))

        from ..utils import _compute_validity_regions
        validity_regions = _compute_validity_regions(series)
        for point in series:
            point.valid_from=validity_regions[point.t][0]
            point.valid_to=validity_regions[point.t][1]

        from ..interpolators import LinearInterpolator
        series_view = _TimeSeriesDenseView(series, 2, 8, dense=True, interpolator_class=LinearInterpolator)

        self.assertEqual(len(series_view), 7)

        series_view_materialized=[]
        for point in series_view:
            series_view_materialized.append(point)

        self.assertEqual(series_view_materialized[0].t, 2)
        self.assertEqual(series_view_materialized[1].t, 3)
        self.assertEqual(series_view_materialized[2].t, 4)
        self.assertEqual(series_view_materialized[3].t, 5.5)
        self.assertEqual(series_view_materialized[3].data['value'], 5.5)
        self.assertEqual(series_view_materialized[4].t, 7)
        self.assertEqual(series_view_materialized[5].t, 8)
        self.assertEqual(series_view_materialized[6].t, 9)


class TestResampler(unittest.TestCase):

    def test_resample_minimal(self):

        # Test series
        series = TimeSeries()
        series.append(DataTimePoint(t=-4,  data={'value':4}))
        series.append(DataTimePoint(t=-3,  data={'value':3}))
        series.append(DataTimePoint(t=-2,  data={'value':2}))
        series.append(DataTimePoint(t=-1,  data={'value':1}))
        series.append(DataTimePoint(t=0,  data={'value':0}))
        series.append(DataTimePoint(t=1,  data={'value':1}))
        series.append(DataTimePoint(t=2,  data={'value':2}))
        series.append(DataTimePoint(t=3,  data={'value':3}))
        series.append(DataTimePoint(t=4,  data={'value':4}))

        # Since by default the resample does not include extremes, we only expect the point
        # at t=0, covering from -2 to +2. Other points would be at t=-4 and t=+4
        resampled_series = series.resample(4)

        # Check len
        self.assertEqual(len(resampled_series), 1)

        # Check the only resampled point
        self.assertEqual(len(resampled_series),1)
        self.assertEqual(resampled_series[0].t,0)
        self.assertEqual(resampled_series[0].data['value'],1)
        self.assertEqual(resampled_series[0].data_loss,0)


    def test_resample_basic(self):

        # Test series
        series = TimeSeries()
        series.append(DataTimePoint(t=-3,  data={'value':3}))
        series.append(DataTimePoint(t=-2,  data={'value':2}))
        series.append(DataTimePoint(t=-1,  data={'value':1}))
        series.append(DataTimePoint(t=0,  data={'value':0}))
        series.append(DataTimePoint(t=1,  data={'value':1}))
        series.append(DataTimePoint(t=2,  data={'value':2}))
        series.append(DataTimePoint(t=3,  data={'value':3}))
        series.append(DataTimePoint(t=4,  data={'value':4}))
        series.append(DataTimePoint(t=5,  data={'value':5}))
        series.append(DataTimePoint(t=6,  data={'value':6}))
        series.append(DataTimePoint(t=7,  data={'value':7}))
        series.append(DataTimePoint(t=16, data={'value':16}))
        series.append(DataTimePoint(t=17, data={'value':17}))
        series.append(DataTimePoint(t=18, data={'value':18}))

        # Resample for 4 seconds
        resampled_series = series.resample(4)
        # DataTimePoint @ 0.0 (1970-01-01 00:00:00+00:00) with data "{'v': 1.0}" and data_loss="0.0"
        # DataTimePoint @ 4.0 (1970-01-01 00:00:04+00:00) with data "{'v': 4.0}" and data_loss="0.0"
        # DataTimePoint @ 8.0 (1970-01-01 00:00:08+00:00) with data "{'v': 7.96875}" and data_loss="0.625"
        # DataTimePoint @ 12.0 (1970-01-01 00:00:12+00:00) with data "{'v': 12.0}" and data_loss="1.0"
        # DataTimePoint @ 16.0 (1970-01-01 00:00:16+00:00) with data "{'v': 16.03125}" and data_loss="0.375"

        # Check len
        self.assertEqual(len(resampled_series), 5)

        # Check for t = 0
        self.assertEqual(resampled_series[0].t, 0)
        self.assertEqual(resampled_series[0].data['value'], 1)  # Expected: 1
        self.assertEqual(resampled_series[0].data_loss, 0)

        # Check for t = 4
        self.assertEqual(resampled_series[1].t, 4)
        self.assertEqual(resampled_series[1].data['value'], 4)  # Expected: 4
        self.assertEqual(resampled_series[1].data_loss, 0)

        # Check for t = 8
        self.assertEqual(resampled_series[2].t, 8)
        self.assertAlmostEqual(resampled_series[2].data['value'], 7.96875)  # Expected: 8
        self.assertAlmostEqual(resampled_series[2].data_loss, 0.625)

        # Check for t = 12
        self.assertEqual(resampled_series[3].t, 12)
        self.assertEqual(resampled_series[3].data['value'], 12)  # Expected: 12 (fully reconstructed)
        self.assertEqual(resampled_series[3].data_loss, 1)

        # Check for t = 16
        self.assertEqual(resampled_series[4].t, 16)
        self.assertAlmostEqual(resampled_series[4].data['value'], 16.03125)  # Expected: 16
        self.assertAlmostEqual(resampled_series[4].data_loss, 0.375)

        # Resample for 3 seconds
        resampled_series = series.resample(3)
        # DataTimePoint @ 0.0 (1970-01-01 00:00:00+00:00) with data "{'value': 0.6666666666666666}" and data_loss="0.0"
        # DataTimePoint @ 3.0 (1970-01-01 00:00:03+00:00) with data "{'value': 3.0}" and data_loss="0.0"
        # DataTimePoint @ 6.0 (1970-01-01 00:00:06+00:00) with data "{'value': 6.0}" and data_loss="0.0"
        # DataTimePoint @ 9.0 (1970-01-01 00:00:09+00:00) with data "{'value': 9.0}" and data_loss="1.0"
        # DataTimePoint @ 12.0 (1970-01-01 00:00:12+00:00) with data "{'value': 12.0}" and data_loss="1.0"
        # DataTimePoint @ 15.0 (1970-01-01 00:00:15+00:00) with data "{'value': 15.0}" and data_loss="0.6666666666666667"

        # Check len
        self.assertEqual(len(resampled_series), 6)

        # Check for t = 0
        self.assertEqual(resampled_series[0].t, 0)
        self.assertAlmostEqual(resampled_series[0].data['value'], 0.6666666666)  # Expected: 0.66..
        self.assertEqual(resampled_series[0].data_loss, 0)

        # Check for t = 3
        self.assertEqual(resampled_series[1].t, 3)
        self.assertEqual(resampled_series[1].data['value'], 3)  # Expected: 3
        self.assertEqual(resampled_series[1].data_loss, 0)

        # Check for t = 6
        self.assertEqual(resampled_series[2].t, 6)
        self.assertEqual(resampled_series[2].data['value'], 6)  # Expected: 6
        self.assertEqual(resampled_series[2].data_loss, 0)

        # Check for t = 9
        self.assertEqual(resampled_series[3].t, 9)
        self.assertEqual(resampled_series[3].data['value'], 9)  # Expected: 9 (fully reconstructed)
        self.assertEqual(resampled_series[3].data_loss, 1)

        # Check for t = 12
        self.assertEqual(resampled_series[4].t, 12)
        self.assertEqual(resampled_series[4].data['value'], 12)  # Expected: 12 (fully reconstructed)
        self.assertEqual(resampled_series[4].data_loss, 1)

        # Check for t = 15
        self.assertEqual(resampled_series[5].t, 15)
        self.assertEqual(resampled_series[5].data['value'], 15)  # Expected: 15
        self.assertAlmostEqual(resampled_series[5].data_loss, 0.6666666666)


    def test_resample_uniform_interpolator(self):

        # Test series
        series = TimeSeries()
        series.append(DataTimePoint(t=-3,  data={'value':3}))
        series.append(DataTimePoint(t=-2,  data={'value':2}))
        series.append(DataTimePoint(t=-1,  data={'value':1}))
        series.append(DataTimePoint(t=0,  data={'value':0}))
        series.append(DataTimePoint(t=1,  data={'value':1}))
        series.append(DataTimePoint(t=2,  data={'value':2}))
        series.append(DataTimePoint(t=3,  data={'value':3}))
        series.append(DataTimePoint(t=4,  data={'value':4}))
        series.append(DataTimePoint(t=5,  data={'value':5}))
        series.append(DataTimePoint(t=6,  data={'value':6}))
        series.append(DataTimePoint(t=7,  data={'value':7}))
        series.append(DataTimePoint(t=16, data={'value':16}))
        series.append(DataTimePoint(t=17, data={'value':17}))
        series.append(DataTimePoint(t=18, data={'value':18}))

        from ..interpolators import UniformInterpolator

        # Resample for 3 seconds
        resampled_series = series.resample(3, interpolator_class=UniformInterpolator)

        # Check len
        self.assertEqual(len(resampled_series), 6)

        # Check for t = 0
        self.assertEqual(resampled_series[0].t, 0)
        self.assertAlmostEqual(resampled_series[0].data['value'], 0.6666666666)
        self.assertEqual(resampled_series[0].data_loss, 0)

        # Check for t = 3
        self.assertEqual(resampled_series[1].t, 3)
        self.assertEqual(resampled_series[1].data['value'], 3)
        self.assertEqual(resampled_series[1].data_loss, 0)

        # Check for t = 6
        self.assertEqual(resampled_series[2].t, 6)
        self.assertEqual(resampled_series[2].data['value'], 6)
        self.assertEqual(resampled_series[2].data_loss, 0)

        # Check for t = 9
        self.assertEqual(resampled_series[3].t, 9)
        self.assertEqual(resampled_series[3].data['value'], 11.5)
        self.assertEqual(resampled_series[3].data_loss, 1)

        # Check for t = 12
        self.assertEqual(resampled_series[4].t, 12)
        self.assertEqual(resampled_series[4].data['value'], 11.5)
        self.assertEqual(resampled_series[4].data_loss, 1)

        # Check for t = 15
        self.assertEqual(resampled_series[5].t, 15)
        self.assertEqual(resampled_series[5].data['value'], 13)
        self.assertAlmostEqual(resampled_series[5].data_loss, 0.6666666666)


    def test_upsample(self):

        # Test series
        series = TimeSeries()
        series.append(DataTimePoint(t=-3,  data={'value':3}))
        series.append(DataTimePoint(t=-2,  data={'value':2}))
        series.append(DataTimePoint(t=-1,  data={'value':1}))
        series.append(DataTimePoint(t=0,  data={'value':0}))
        series.append(DataTimePoint(t=1,  data={'value':1}))
        series.append(DataTimePoint(t=2,  data={'value':2}))
        series.append(DataTimePoint(t=3,  data={'value':3}))
        series.append(DataTimePoint(t=4,  data={'value':4}))
        series.append(DataTimePoint(t=5,  data={'value':5}))
        series.append(DataTimePoint(t=6,  data={'value':6}))
        series.append(DataTimePoint(t=7,  data={'value':7}))
        series.append(DataTimePoint(t=16, data={'value':16}))
        series.append(DataTimePoint(t=17, data={'value':17}))
        series.append(DataTimePoint(t=18, data={'value':18}))

        # Resample (upsample) for 0.5 seconds
        upsampled_series = series.resample(0.5)

        # Check start/end timestamps
        self.assertEqual(upsampled_series[0].t, -2.5)
        self.assertEqual(upsampled_series[-1].t, 17.5)

        # Check values & data losses
        self.assertEqual(upsampled_series[19].data['value'], 7)
        self.assertEqual(upsampled_series[19].data_loss, 0)

        self.assertEqual(upsampled_series[20].data['value'], 7.3125)
        # TODO: the above should be 7.5, fix me! See also test_upgragate
        self.assertEqual(upsampled_series[20].data_loss, 0.5)

        self.assertEqual(upsampled_series[21].data['value'], 8)
        self.assertEqual(upsampled_series[21].data_loss, 1)


    def test_resample_edge_1(self):

        # Create 5-minute test data with
        series = TimeSeries()
        series.append(DataTimePoint(t = (60*20), data = {'value': 1.55}))
        series.append(DataTimePoint(t = (60*25), data = {'value': 1.58}))
        series.append(DataTimePoint(t = (60*30), data = {'value': 1.61}))
        series.append(DataTimePoint(t = (60*45), data = {'value': 1.70}))
        series.append(DataTimePoint(t = (60*50), data = {'value': 1.73}))

        # Resample
        resampled_series = Resampler('300s').process(series)

        # Check len
        self.assertEqual(len(resampled_series), 5)

        # Do not check for t = 20 as extremes are not included

        # Check for t = 25
        self.assertEqual(resampled_series[0].t, (60*25))
        self.assertAlmostEqual(resampled_series[0].data['value'], 1.58)
        self.assertEqual(resampled_series[0].data_loss, 0)

        # Check for t = 30
        self.assertEqual(resampled_series[1].t, (60*30))
        self.assertEqual(resampled_series[1].data['value'], 1.61)
        self.assertEqual(resampled_series[1].data_loss, 0)

        # Check for t = 35
        self.assertEqual(resampled_series[2].t, (60*35))
        self.assertAlmostEqual(resampled_series[2].data['value'], 1.64)
        self.assertEqual(resampled_series[2].data_loss, 1)

        # Check for t = 40
        self.assertEqual(resampled_series[3].t, (60*40))
        self.assertAlmostEqual(resampled_series[3].data['value'], 1.67)
        self.assertEqual(resampled_series[3].data_loss, 1)

        # Check for t = 45
        self.assertEqual(resampled_series[4].t, (60*45))
        self.assertAlmostEqual(resampled_series[4].data['value'], 1.70)
        self.assertEqual(resampled_series[4].data_loss, 0)

        # Do not check for t = 50 as extremes are not included


    def test_resample_edge_2(self):

        # Create 5-minute test data  with ten secs ofset
        series = TimeSeries()
        series.append(DataTimePoint(t = (60*20)+10, data = {'value': 1.55}))
        series.append(DataTimePoint(t = (60*25)+10, data = {'value': 1.58}))
        series.append(DataTimePoint(t = (60*30)+10, data = {'value': 1.61}))
        series.append(DataTimePoint(t = (60*45)+10, data = {'value': 1.71}))
        series.append(DataTimePoint(t = (60*50)+10, data = {'value': 1.74}))

        # Resample
        resampled_series = Resampler('300s').process(series)

        # Check len
        self.assertEqual(len(resampled_series), 5)

        # Do not check for t = 20 as extremes are not included

        # Check for t = 25
        self.assertEqual(resampled_series[0].t, (60*25))
        self.assertAlmostEqual(resampled_series[0].data['value'], 1.579)
        self.assertEqual(resampled_series[0].data_loss, 0)

        # Check for t = 30
        self.assertEqual(resampled_series[1].t, (60*30))
        self.assertEqual(resampled_series[1].data['value'], 1.609)
        self.assertEqual(resampled_series[1].data_loss, 0)

        # Check for t = 35
        self.assertEqual(resampled_series[2].t, (60*35))
        self.assertAlmostEqual(resampled_series[2].data['value'], 1.642, 3)
        self.assertAlmostEqual(resampled_series[2].data_loss, 290/300)

        # Check for t = 40
        self.assertEqual(resampled_series[3].t, (60*40))
        self.assertAlmostEqual(resampled_series[3].data['value'], 1.676, 3)
        self.assertEqual(resampled_series[3].data_loss, 1)

        # Check for t = 45
        self.assertEqual(resampled_series[4].t, (60*45))
        self.assertAlmostEqual(resampled_series[4].data['value'], 1.709, 3)
        self.assertAlmostEqual(resampled_series[4].data_loss, 10/300)

        # Do not check for t = 50 as extremes are not included


    def test_resampler_data_indexes(self):

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        series = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i < 15:
                anomaly = 0.25
                forecast = 0
                data_reconstructed = 1
                data_loss = 1
            elif i > 20:
                anomaly = None
                forecast = 1
                data_reconstructed = 0
                data_loss = None
            else:
                if i>10:
                    anomaly = 0.7
                else:
                    anomaly = 0
                forecast = 0
                data_reconstructed = 0
                data_loss = 0.1

            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i, 'humidity': 5},
                                  data_loss = data_loss)

            # Add extra data_indexes as if something operated on the time series
            point.data_indexes['data_reconstructed'] = data_reconstructed
            point.data_indexes['anomaly'] = anomaly
            point.data_indexes['forecast'] = forecast

            # Append
            series.append(point)

        # Resample the time series
        resampled_series = Resampler('600s').process(series)

        # Check that we still have all of  the data_indexes
        self.assertEqual(resampled_series._all_data_indexes(), ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])

        # Check data_indexes math
        self.assertAlmostEqual(resampled_series[0].data_indexes['data_reconstructed'], 0.4)
        self.assertAlmostEqual(resampled_series[0].data_indexes['anomaly'], 0.275)
        self.assertAlmostEqual(resampled_series[0].data_indexes['forecast'], 0)
        self.assertAlmostEqual(resampled_series[0].data_loss, 0.46)

        self.assertAlmostEqual(resampled_series[1].data_indexes['data_reconstructed'], 0)
        self.assertAlmostEqual(resampled_series[1].data_indexes['anomaly'], 0.7)
        self.assertAlmostEqual(resampled_series[1].data_indexes['forecast'], 0.65)
        self.assertAlmostEqual(resampled_series[1].data_loss, 0.035)

        # Lastly, resample at the same sampling interval and check that all data_indexes are still there as well
        same_resampled_series = series[0:5].resample(60)
        self.assertEqual(same_resampled_series._all_data_indexes(), ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])



class TestAggregator(unittest.TestCase):

    def setUp(self):

        # Time series at 1m resolution from 14:58:00 to 15:32:00 (UTC)
        self.series_1 = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i})
            self.series_1.append(point)

        # Time series at 1m resolution  from 14:00:00 to 15:30:00 (UTC)
        self.series_2 = TimeSeries()
        start_t = 1436022000
        for i in range(34):
            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i})
            self.series_2.append(point)

        # Time series at 1m resolution from 14:00:00 to 13:20:00 (UTC)
        self.series_3 = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(23):
            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i})
            self.series_3.append(point)

        # Time series at 1m resolution from 14:10:00 to 15:30:00 (UTC)
        self.series_4 = TimeSeries()
        start_t = 1436022000 + 600
        for i in range(21):
            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i})
            self.series_4.append(point)

        # Time series at 1m resolution from 14:58:00 to 15:32:00 (UTC)
        self.series_5 = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i <21:
                continue
            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i})
            self.series_5.append(point)

        # Time series at 15m resolution from 01:00 to 06:00 (Europe/Rome)
        from_dt  = dt(2019,10,1,1,0,0, tz='Europe/Rome')
        to_dt    = dt(2019,10,1,6,0,0, tz='Europe/Rome')
        time_unit = TimeUnit('15m')
        self.series_6 = TimeSeries()
        slider_dt = from_dt
        count = 0
        while slider_dt < to_dt:
            if count not in [1, 6, 7, 8, 9, 10]:
                point = DataTimePoint(t = s_from_dt(slider_dt),
                                      data = {'temperature': 154+count},
                                      tz = 'Europe/Rome')
                self.series_6.append(point)
            slider_dt = slider_dt + time_unit
            count += 1

        # Time series at 1h resolution from 2019-10-24 00:00 to 2019-10-31 00:00:(Europe/Rome), DST off -> 2 AM repeated
        from_dt = dt(2019,10,24,0,0,0, tz='Europe/Rome')
        to_dt   = dt(2019,10,31,0,0,0, tz='Europe/Rome')
        time_unit = TimeUnit('1h')
        self.series_7 = TimeSeries()
        slider_dt = from_dt
        count = 0
        while slider_dt < to_dt:
            if count not in []:
                point = DataTimePoint(t = s_from_dt(slider_dt),
                                      data = {'temperature': 154+count},
                                      tz = 'Europe/Rome')
                self.series_7.append(point)
            slider_dt = slider_dt + time_unit
            count += 1


    def test_aggregate(self):

        # Time series at 1m resolution from 14:58:00 to 15:32:00 (UTC), no from/to
        series = Aggregator('600s').process(self.series_1)
        self.assertEqual(len(series), 3)
        self.assertEqual(series[0].start.dt, dt_from_str('2015-07-04 15:00:00+00:00'))
        self.assertEqual(series[1].start.dt, dt_from_str('2015-07-04 15:10:00+00:00'))
        self.assertEqual(series[2].start.dt, dt_from_str('2015-07-04 15:20:00+00:00'))

        # Time series at 15m resolution from 01:00 to 06:00 (Europe/Rome), slot in 1h
        series = Aggregator('1h').process(self.series_6)
        self.assertEqual(len(series), 4)
        self.assertEqual(str(series[0].start.dt), str('2019-10-01 01:00:00+02:00'))
        self.assertEqual(str(series[1].start.dt), str('2019-10-01 02:00:00+02:00'))
        self.assertEqual(str(series[2].start.dt), str('2019-10-01 03:00:00+02:00'))
        self.assertEqual(str(series[3].start.dt), str('2019-10-01 04:00:00+02:00'))

        # Test with changing time zone of the series
        self.series_6.change_tz('America/New_York')
        series = Aggregator('1h').process(self.series_6)
        self.assertEqual(len(series), 4)
        self.assertEqual(str(series[0].start.dt), str('2019-09-30 19:00:00-04:00'))
        self.assertEqual(str(series[1].start.dt), str('2019-09-30 20:00:00-04:00'))
        self.assertEqual(str(series[2].start.dt), str('2019-09-30 21:00:00-04:00'))
        self.assertEqual(str(series[3].start.dt), str('2019-09-30 22:00:00-04:00'))

        # Time series from 2019,10,24,0,0,0 to 2019,10,31,0,0,0 (Europe/Rome), DST off -> 2 AM repeated
        series = Aggregator('1h').process(self.series_7)
        self.assertEqual(len(series), 168)
        self.assertEqual(str(series[73].start.dt), str('2019-10-27 01:00:00+02:00'))
        self.assertEqual(str(series[74].start.dt), str('2019-10-27 02:00:00+02:00'))
        self.assertEqual(str(series[75].start.dt), str('2019-10-27 02:00:00+01:00'))
        self.assertEqual(str(series[76].start.dt), str('2019-10-27 03:00:00+01:00'))
        self.assertEqual(series[76].data['temperature_avg'],230.5)

        # Time series from 2019,10,24,0,0,0 to 2019,10,31,0,0,0 (Europe/Rome), DST off -> 2 AM repeated
        series = Aggregator('1D').process(self.series_7)
        self.assertEqual(len(series), 6)
        self.assertEqual(str(series[0].start.dt), str('2019-10-24 00:00:00+02:00'))
        self.assertEqual(str(series[-1].start.dt), str('2019-10-29 00:00:00+01:00')) # Last one not included as right excluded


    def test_upgregate(self):

        series = TimeSeries()
        series.append(DataTimePoint(t = 0, data = {'value': 0}))
        series.append(DataTimePoint(t = 3600, data = {'value': 3600}))
        series.append(DataTimePoint(t = 7200, data = {'value': 7200}))

        # This is an upgregate (similar to an upsamplig)
        aggregator = Aggregator('10m')
        aggregated_series = aggregator.process(series)
        self.assertEqual(len(aggregated_series), 12)

        # TODO: check the values and the math. See also test_upsample


    def test_aggregate_operations_and_losses(self):

        # Import (overloaded) operations
        from ..operations import min, max, avg, sum

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        series = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i, 'humidity': 5})
            series.append(point)

        # Add extra operations
        slotted_series = Aggregator('600s', operations=[avg,min,max]).process(series)
        self.assertAlmostEqual(slotted_series[0].data['temperature_min'], 156)
        self.assertAlmostEqual(slotted_series[0].data['temperature_max'], 165)
        self.assertAlmostEqual(slotted_series[0].data['temperature_avg'], 161)
        self.assertAlmostEqual(slotted_series[0].data['humidity_min'], 5)
        self.assertAlmostEqual(slotted_series[0].data['humidity_max'], 5)
        self.assertAlmostEqual(slotted_series[0].data['humidity_avg'], 5)

        # Entirely change the operation
        slotted_series = Aggregator('600s', operations=[sum]).process(series)
        self.assertEqual(slotted_series[0].data['temperature_sum'], 1605) # 156+157+158+159+160+161+162+163+164+165
        self.assertEqual(slotted_series[0].data['humidity_sum'], 50)
        self.assertEqual(slotted_series[1].data['temperature_sum'], 1705)
        self.assertEqual(slotted_series[1].data['humidity_sum'], 50)

        # Operations as strings
        slotted_series = Aggregator('600s', operations=['avg','min',max]).process(series)
        self.assertAlmostEqual(slotted_series[0].data['temperature_min'], 156)
        self.assertAlmostEqual(slotted_series[0].data['temperature_max'], 165)
        self.assertAlmostEqual(slotted_series[0].data['temperature_avg'], 161)
        self.assertAlmostEqual(slotted_series[0].data['humidity_min'], 5)
        self.assertAlmostEqual(slotted_series[0].data['humidity_max'], 5)
        self.assertAlmostEqual(slotted_series[0].data['humidity_avg'], 5)

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome) with a full short (single-slot) gap to be interpolated
        series = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i < 12 or i > 22:
                point = DataTimePoint(t = start_t + (i*60),
                                      data = {'temperature': 154+i, 'humidity': 5})
                series.append(point)

        slotted_series = Aggregator('600s', operations=[avg,min,max]).process(series)
        self.assertEqual(len(slotted_series), 3)

        self.assertAlmostEqual(slotted_series[0].data['humidity_min'], 5)
        self.assertAlmostEqual(slotted_series[0].data['humidity_max'], 5)
        self.assertAlmostEqual(slotted_series[0].data['humidity_avg'], 5)
        self.assertAlmostEqual(slotted_series[0].data['temperature_min'], 156)
        self.assertAlmostEqual(slotted_series[0].data['temperature_max'], 165)
        self.assertAlmostEqual(slotted_series[0].data['temperature_avg'], 160.9875)

        self.assertAlmostEqual(slotted_series[1].data['humidity_min'], 5)
        self.assertAlmostEqual(slotted_series[1].data['humidity_max'], 5)
        self.assertAlmostEqual(slotted_series[1].data['humidity_avg'], 5)
        self.assertAlmostEqual(slotted_series[1].data['temperature_min'], 171)
        self.assertAlmostEqual(slotted_series[1].data['temperature_max'], 171)
        self.assertAlmostEqual(slotted_series[1].data['temperature_avg'], 171)

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome) with a full long (multi-slot) gap to be interpolated
        series = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(60):
            if i < 10 or i >48:
                point = DataTimePoint(t = start_t + (i*60),
                                      data = {'temperature': 154+i, 'humidity': 5})
                series.append(point)

        slotted_series = Aggregator('600s', operations=[avg,min,max]).process(series)
        self.assertEqual(len(slotted_series), 5)

        self.assertEqual(slotted_series[0].data['temperature_min'], 156)
        self.assertEqual(slotted_series[0].data['temperature_avg'], 160.9875)
        self.assertEqual(slotted_series[1].data['temperature_min'], 171.0)
        self.assertEqual(slotted_series[1].data['temperature_avg'], 171.0)
        self.assertEqual(slotted_series[3].data['temperature_min'], 191.0)
        self.assertEqual(slotted_series[3].data['temperature_avg'], 191.0)


    def test_aggregate_data_indexes(self):

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        series = TimeSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i < 15:
                anomaly = 0.25
                forecast = 0
                data_reconstructed = 1
                data_loss = 1
            elif i > 20:
                anomaly = None
                forecast = 1
                data_reconstructed = 0
                data_loss = None
            else:
                if i>10:
                    anomaly = 0.7
                else:
                    anomaly = 0
                forecast = 0
                data_reconstructed = 0
                data_loss = 0.1

            point = DataTimePoint(t = start_t + (i*60),
                                  data = {'temperature': 154+i, 'humidity': 5},
                                  data_loss = data_loss)

            # Add extra data_indexes as if something operated on the time series
            point.data_indexes['data_reconstructed'] = data_reconstructed
            point.data_indexes['anomaly'] = anomaly
            point.data_indexes['forecast'] = forecast

            # Append
            series.append(point)

        # Slot the time series
        slotted_series = Aggregator('600s').process(series)

        # Check that we have all the data_indexes
        self.assertEqual(slotted_series._all_data_indexes(), ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])

        # Check data_indexes math
        self.assertAlmostEqual(slotted_series[0].data_indexes['anomaly'], 0.0375)
        self.assertAlmostEqual(slotted_series[0].data_indexes['data_reconstructed'], 0.15)
        self.assertAlmostEqual(slotted_series[0].data_indexes['forecast'], 0)

        self.assertAlmostEqual(slotted_series[1].data_indexes['anomaly'], 0.5676, 4)
        self.assertAlmostEqual(slotted_series[1].data_indexes['data_reconstructed'], 0.25)
        self.assertAlmostEqual(slotted_series[1].data_indexes['forecast'], 0.15)

        self.assertEqual(slotted_series[2].data_indexes['anomaly'], None)
        self.assertAlmostEqual(slotted_series[2].data_indexes['data_reconstructed'], 0)
        self.assertEqual(slotted_series[2].data_indexes['forecast'], 1)

        # The following is re-computed from missing coverage only, and not marked as None.
        # See the compute_data_loss function fore more details and some more comments.
        self.assertEqual(slotted_series[2].data_loss, 0)


